(*
* Copyright (c) 2015, Facebook, Inc.
* All rights reserved.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the "hack" directory of this source tree.
*
*)

open Hh_prelude
open Decl_defs
open Shallow_decl_defs
open Typing_defs

(*****************************************************************************)
(* Functor traversing a type, but applies a user defined function for
 * positions. Positions in errors (_decl_errors) are not mapped - entire
 * field is erased instead. This is safe because there exists a completely
 * different system for tracking and updating positions in errors
 * (see ServerTypeCheck.get_files_with_stale_errors)
 *)
(*****************************************************************************)
module TraversePos (ImplementPos : sig
  val pos : Pos.t -> Pos.t
end) =
struct
  open Typing_reason

  let pos = ImplementPos.pos

  let string_id (p, x) = (pos p, x)

  let rec reason = function
    | Rnone -> Rnone
    | Rwitness p -> Rwitness (pos p)
    | Ridx (p, r) -> Ridx (pos p, reason r)
    | Ridx_vector p -> Ridx_vector (pos p)
    | Rforeach p -> Rforeach (pos p)
    | Rasyncforeach p -> Rasyncforeach (pos p)
    | Rarith p -> Rarith (pos p)
    | Rarith_ret p -> Rarith_ret (pos p)
    | Rcomp p -> Rcomp (pos p)
    | Rconcat_ret p -> Rconcat_ret (pos p)
    | Rlogic_ret p -> Rlogic_ret (pos p)
    | Rbitwise p -> Rbitwise (pos p)
    | Rbitwise_ret p -> Rbitwise_ret (pos p)
    | Rno_return p -> Rno_return (pos p)
    | Rno_return_async p -> Rno_return_async (pos p)
    | Rret_fun_kind (p, k) -> Rret_fun_kind (pos p, k)
    | Rhint p -> Rhint (pos p)
    | Rthrow p -> Rthrow (pos p)
    | Rplaceholder p -> Rplaceholder (pos p)
    | Rret_div p -> Rret_div (pos p)
    | Ryield_gen p -> Ryield_gen (pos p)
    | Ryield_asyncgen p -> Ryield_asyncgen (pos p)
    | Ryield_asyncnull p -> Ryield_asyncnull (pos p)
    | Ryield_send p -> Ryield_send (pos p)
    | Rlost_info (s, r1, Blame (p2, l)) ->
      Rlost_info (s, reason r1, Blame (pos p2, l))
    | Rformat (p1, s, r) -> Rformat (pos p1, s, reason r)
    | Rclass_class (p, s) -> Rclass_class (pos p, s)
    | Runknown_class p -> Runknown_class (pos p)
    | Rvar_param p -> Rvar_param (pos p)
    | Runpack_param (p1, p2, i) -> Runpack_param (pos p1, pos p2, i)
    | Rinout_param p -> Rinout_param (pos p)
    | Rinstantiate (r1, x, r2) -> Rinstantiate (reason r1, x, reason r2)
    | Rarray_filter (p, r) -> Rarray_filter (pos p, reason r)
    | Rtypeconst (r1, (p, s1), s2, r2) ->
      Rtypeconst (reason r1, (pos p, s1), s2, reason r2)
    | Rtype_access (r1, ls) ->
      Rtype_access (reason r1, List.map ls ~f:(fun (r, s) -> (reason r, s)))
    | Rexpr_dep_type (r, p, n) -> Rexpr_dep_type (reason r, pos p, n)
    | Rnullsafe_op p -> Rnullsafe_op (pos p)
    | Rtconst_no_cstr (p, s) -> Rtconst_no_cstr (pos p, s)
    | Rpredicated (p, f) -> Rpredicated (pos p, f)
    | Ris p -> Ris (pos p)
    | Ras p -> Ras (pos p)
    | Rvarray_or_darray_key p -> Rvarray_or_darray_key (pos p)
    | Rusing p -> Rusing (pos p)
    | Rdynamic_prop p -> Rdynamic_prop (pos p)
    | Rdynamic_call p -> Rdynamic_call (pos p)
    | Rdynamic_construct p -> Rdynamic_construct (pos p)
    | Ridx_dict p -> Ridx_dict (pos p)
    | Rmissing_required_field (p, n) -> Rmissing_required_field (pos p, n)
    | Rmissing_optional_field (p, n) -> Rmissing_optional_field (pos p, n)
    | Runset_field (p, n) -> Runset_field (pos p, n)
    | Rcontravariant_generic (r1, n) -> Rcontravariant_generic (reason r1, n)
    | Rinvariant_generic (r1, n) -> Rcontravariant_generic (reason r1, n)
    | Rregex p -> Rregex (pos p)
    | Rimplicit_upper_bound (p, s) -> Rimplicit_upper_bound (pos p, s)
    | Rarith_ret_int p -> Rarith_ret_int (pos p)
    | Rarith_ret_float (p, r, s) -> Rarith_ret_float (pos p, reason r, s)
    | Rarith_ret_num (p, r, s) -> Rarith_ret_num (pos p, reason r, s)
    | Rarith_dynamic p -> Rarith_dynamic (pos p)
    | Rbitwise_dynamic p -> Rbitwise_dynamic (pos p)
    | Rincdec_dynamic p -> Rincdec_dynamic (pos p)
    | Rtype_variable p -> Rtype_variable (pos p)
    | Rtype_variable_generics (p, t, s) -> Rtype_variable_generics (pos p, t, s)
    | Rsolve_fail p -> Rsolve_fail (pos p)
    | Rcstr_on_generics (p, sid) -> Rcstr_on_generics (pos p, string_id sid)
    | Rlambda_param (p, r) -> Rlambda_param (pos p, reason r)
    | Rshape (p, fun_name) -> Rshape (pos p, fun_name)
    | Renforceable p -> Renforceable (pos p)
    | Rdestructure p -> Rdestructure (pos p)
    | Rkey_value_collection_key p -> Rkey_value_collection_key (pos p)
    | Rglobal_class_prop p -> Rglobal_class_prop (pos p)
    | Rglobal_fun_param p -> Rglobal_fun_param (pos p)
    | Rglobal_fun_ret p -> Rglobal_fun_ret (pos p)
    | Rsplice p -> Rsplice (pos p)
    | Ret_boolean p -> Ret_boolean (pos p)
    | Rdefault_capability p -> Rdefault_capability (pos p)

  let rec ty t =
    let (p, x) = deref t in
    mk (reason p, ty_ x)

  and ty_ : decl_phase ty_ -> decl_phase ty_ = function
    | (Tany _ | Tthis | Terr | Tmixed | Tnonnull | Tdynamic | Tvar _) as x -> x
    | Tdarray (ty1, ty2) -> Tdarray (ty ty1, ty ty2)
    | Tvarray root_ty -> Tvarray (ty root_ty)
    | Tvarray_or_darray (ty1, ty2) -> Tvarray_or_darray (ty ty1, ty ty2)
    | Tprim _ as x -> x
    | Tgeneric (name, args) -> Tgeneric (name, List.map args ty)
    | Ttuple tyl -> Ttuple (List.map tyl ty)
    | Tunion tyl -> Tunion (List.map tyl ty)
    | Tintersection tyl -> Tintersection (List.map tyl ty)
    | Toption x -> Toption (ty x)
    | Tlike x -> Tlike (ty x)
    | Tfun ft -> Tfun (fun_type ft)
    | Tapply (sid, xl) -> Tapply (string_id sid, List.map xl ty)
    | Taccess (root_ty, id) -> Taccess (ty root_ty, string_id id)
    | Tshape (shape_kind, fdm) ->
      Tshape (shape_kind, ShapeFieldMap.map_and_rekey fdm shape_field_name ty)

  and ty_opt x = Option.map x ty

  and shape_field_name = function
    | Ast_defs.SFlit_int s -> Ast_defs.SFlit_int (string_id s)
    | Ast_defs.SFlit_str s -> Ast_defs.SFlit_str (string_id s)
    | Ast_defs.SFclass_const (id, s) ->
      Ast_defs.SFclass_const (string_id id, string_id s)

  and constraint_ x = List.map ~f:(fun (ck, x) -> (ck, ty x)) x

  and possibly_enforced_ty et = { et with et_type = ty et.et_type }

  and capability = function
    | CapTy cap -> CapTy (ty cap)
    | CapDefaults p -> CapDefaults (pos p)

  and fun_implicit_params implicit =
    { capability = capability implicit.capability }

  and fun_type ft =
    {
      ft with
      ft_tparams = List.map ~f:type_param ft.ft_tparams;
      ft_where_constraints = List.map ft.ft_where_constraints where_constraint;
      ft_params = List.map ft.ft_params fun_param;
      ft_implicit_params = fun_implicit_params ft.ft_implicit_params;
      ft_ret = possibly_enforced_ty ft.ft_ret;
      ft_arity = fun_arity ft.ft_arity;
      ft_reactive = fun_reactive ft.ft_reactive;
    }

  and fun_elt fe = { fe with fe_type = ty fe.fe_type; fe_pos = pos fe.fe_pos }

  and fun_reactive = function
    | Local (Some ty1) -> Local (Some (ty ty1))
    | Shallow (Some ty1) -> Shallow (Some (ty ty1))
    | Reactive (Some ty1) -> Reactive (Some (ty ty1))
    | Pure (Some ty1) -> Pure (Some (ty ty1))
    | r -> r

  and where_constraint (ty1, c, ty2) = (ty ty1, c, ty ty2)

  and fun_arity = function
    | Fstandard as x -> x
    | Fvariadic param -> Fvariadic (fun_param param)

  and fun_param param =
    {
      param with
      fp_pos = pos param.fp_pos;
      fp_type = possibly_enforced_ty param.fp_type;
      fp_rx_annotation = param_rx_annotation param.fp_rx_annotation;
    }

  and param_rx_annotation = function
    | Some (Param_rx_if_impl t) -> Some (Param_rx_if_impl (ty t))
    | c -> c

  and class_const cc =
    {
      cc_synthesized = cc.cc_synthesized;
      cc_abstract = cc.cc_abstract;
      cc_pos = pos cc.cc_pos;
      cc_type = ty cc.cc_type;
      cc_origin = cc.cc_origin;
    }

  and typeconst_abstract_kind = function
    | TCAbstract default -> TCAbstract (ty_opt default)
    | TCPartiallyAbstract -> TCPartiallyAbstract
    | TCConcrete -> TCConcrete

  and typeconst tc =
    {
      ttc_abstract = typeconst_abstract_kind tc.ttc_abstract;
      ttc_name = string_id tc.ttc_name;
      ttc_constraint = ty_opt tc.ttc_constraint;
      ttc_type = ty_opt tc.ttc_type;
      ttc_origin = tc.ttc_origin;
      ttc_enforceable = Tuple.T2.map_fst ~f:pos tc.ttc_enforceable;
      ttc_reifiable = Option.map tc.ttc_reifiable pos;
    }

  and user_attribute { ua_name; ua_classname_params } =
    { ua_name = string_id ua_name; ua_classname_params }

  and type_param t =
    {
      tp_name = string_id t.tp_name;
      tp_variance = t.tp_variance;
      tp_reified = t.tp_reified;
      tp_tparams = List.map ~f:type_param t.tp_tparams;
      tp_constraints = constraint_ t.tp_constraints;
      tp_user_attributes = List.map ~f:user_attribute t.tp_user_attributes;
    }

  and class_type dc =
    {
      dc_final = dc.dc_final;
      dc_const = dc.dc_const;
      dc_need_init = dc.dc_need_init;
      dc_deferred_init_members = dc.dc_deferred_init_members;
      dc_abstract = dc.dc_abstract;
      dc_members_fully_known = dc.dc_members_fully_known;
      dc_kind = dc.dc_kind;
      dc_is_xhp = dc.dc_is_xhp;
      dc_has_xhp_keyword = dc.dc_has_xhp_keyword;
      dc_is_disposable = dc.dc_is_disposable;
      dc_name = dc.dc_name;
      dc_pos = dc.dc_pos;
      dc_extends = dc.dc_extends;
      dc_sealed_whitelist = dc.dc_sealed_whitelist;
      dc_xhp_attr_deps = dc.dc_xhp_attr_deps;
      dc_req_ancestors = List.map dc.dc_req_ancestors requirement;
      dc_req_ancestors_extends = dc.dc_req_ancestors_extends;
      dc_tparams = List.map dc.dc_tparams type_param;
      dc_where_constraints = List.map dc.dc_where_constraints where_constraint;
      dc_substs =
        SMap.map
          begin
            fun ({ sc_subst; _ } as sc) ->
            { sc with sc_subst = SMap.map ty sc_subst }
          end
          dc.dc_substs;
      dc_consts = SMap.map class_const dc.dc_consts;
      dc_typeconsts = SMap.map typeconst dc.dc_typeconsts;
      dc_props = dc.dc_props;
      dc_sprops = dc.dc_sprops;
      dc_methods = dc.dc_methods;
      dc_smethods = dc.dc_smethods;
      dc_construct = dc.dc_construct;
      dc_ancestors = SMap.map ty dc.dc_ancestors;
      dc_implements_dynamic = dc.dc_implements_dynamic;
      dc_enum_type = Option.map dc.dc_enum_type enum_type;
      dc_decl_errors = None;
      dc_condition_types = dc.dc_condition_types;
    }

  and requirement (p, t) = (pos p, ty t)

  and enum_type te =
    {
      te_base = ty te.te_base;
      te_constraint = ty_opt te.te_constraint;
      te_includes = List.map te.te_includes ty;
      te_enum_class = te.te_enum_class;
    }

  and typedef tdef =
    {
      td_pos = pos tdef.td_pos;
      td_vis = tdef.td_vis;
      td_tparams = List.map tdef.td_tparams type_param;
      td_constraint = ty_opt tdef.td_constraint;
      td_type = ty tdef.td_type;
    }

  and shallow_class sc =
    {
      sc_mode = sc.sc_mode;
      sc_final = sc.sc_final;
      sc_is_xhp = sc.sc_is_xhp;
      sc_has_xhp_keyword = sc.sc_has_xhp_keyword;
      sc_kind = sc.sc_kind;
      sc_name = string_id sc.sc_name;
      sc_tparams = List.map sc.sc_tparams type_param;
      sc_where_constraints = List.map sc.sc_where_constraints where_constraint;
      sc_extends = List.map sc.sc_extends ty;
      sc_uses = List.map sc.sc_uses ty;
      sc_xhp_attr_uses = List.map sc.sc_xhp_attr_uses ty;
      sc_req_extends = List.map sc.sc_req_extends ty;
      sc_req_implements = List.map sc.sc_req_implements ty;
      sc_implements = List.map sc.sc_implements ty;
      sc_implements_dynamic = sc.sc_implements_dynamic;
      sc_consts = List.map sc.sc_consts shallow_class_const;
      sc_typeconsts = List.map sc.sc_typeconsts shallow_typeconst;
      sc_props = List.map sc.sc_props shallow_prop;
      sc_sprops = List.map sc.sc_sprops shallow_prop;
      sc_constructor = Option.map sc.sc_constructor shallow_method;
      sc_static_methods = List.map sc.sc_static_methods shallow_method;
      sc_methods = List.map sc.sc_methods shallow_method;
      sc_user_attributes = List.map sc.sc_user_attributes user_attribute;
      sc_enum_type = Option.map sc.sc_enum_type enum_type;
    }

  and shallow_class_const scc =
    {
      scc_abstract = scc.scc_abstract;
      scc_name = string_id scc.scc_name;
      scc_type = ty scc.scc_type;
    }

  and shallow_typeconst stc =
    {
      stc_abstract = typeconst_abstract_kind stc.stc_abstract;
      stc_constraint = Option.map stc.stc_constraint ty;
      stc_name = string_id stc.stc_name;
      stc_type = Option.map stc.stc_type ty;
      stc_enforceable = (pos (fst stc.stc_enforceable), snd stc.stc_enforceable);
      stc_reifiable = Option.map stc.stc_reifiable pos;
    }

  and shallow_prop sp =
    {
      sp_name = string_id sp.sp_name;
      sp_xhp_attr = sp.sp_xhp_attr;
      sp_type = Option.map sp.sp_type ty;
      sp_visibility = sp.sp_visibility;
      sp_flags = sp.sp_flags;
    }

  and shallow_method sm =
    {
      sm_name = string_id sm.sm_name;
      sm_reactivity = sm.sm_reactivity;
      sm_type = ty sm.sm_type;
      sm_visibility = sm.sm_visibility;
      sm_deprecated = sm.sm_deprecated;
      sm_flags = sm.sm_flags;
    }
end

(*****************************************************************************)
(* Returns a signature with all the positions replaced with Pos.none *)
(*****************************************************************************)
module NormalizeSig = TraversePos (struct
  let pos _ = Pos.none
end)
